# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.rgb_ring import (
    RgbRingAction,
    RgbRingBlankAction,
    RgbRingBlinkAction,
    RgbRingColorfulCommandAction,
    RgbRingCommandAction,
    RgbRingCommandEvent,
    RgbRingFillDownfromAction,
    RgbRingFillUptoAction,
    RgbRingProgressWheelAction,
    RgbRingProgressWheelStepAction,
    RgbRingPulseAction,
    RgbRingRainbowAction,
    RgbRingSetAllAction,
    RgbRingSetBrightnessAction,
    RgbRingSetEnabledAction,
    RgbRingSetIsBusyAction,
    RgbRingSetIsConnectedAction,
    RgbRingSpinningWheelAction,
    RgbRingState,
)

Action = InitAction | RgbRingAction


def reducer(  # noqa: C901, PLR0912
    state: RgbRingState | None,
    action: Action,
) -> ReducerResult[RgbRingState, Action, RgbRingCommandEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return RgbRingState(is_connected=False, is_busy=False)
        raise InitializationActionError(action)

    if isinstance(action, RgbRingSetIsConnectedAction):
        return replace(
            state,
            is_connected=action.is_connected,
        )

    if isinstance(action, RgbRingSetIsBusyAction):
        return replace(
            state,
            is_busy=action.is_busy,
        )

    if isinstance(action, RgbRingCommandAction):
        if not state.is_connected:
            return state

        command = None

        if isinstance(action, RgbRingColorfulCommandAction):
            (r, g, b, *_) = action.color
            if isinstance(action, RgbRingSetAllAction):
                command = f'set_all {r} {g} {b}'
            elif isinstance(action, RgbRingBlinkAction):
                command = f'blink {r} {g} {b} {action.wait} {action.repetitions}'
            elif isinstance(action, RgbRingProgressWheelStepAction):
                command = f'progress_wheel_step {r} {g} {b}'
            elif isinstance(action, RgbRingPulseAction):
                command = f'pulse {r} {g} {b} {action.wait} {action.repetitions}'
            elif isinstance(action, RgbRingSpinningWheelAction):
                command = f"""spinning_wheel {r} {g} {b} {action.wait} {action.length} {
                action.repetitions}"""
            elif isinstance(action, RgbRingProgressWheelAction):
                command = f'progress_wheel {r} {g} {b} {action.percentage}'
            elif isinstance(action, RgbRingFillUptoAction):
                command = f'fill_upto {r} {g} {b} {action.percentage} {action.wait}'
            elif isinstance(action, RgbRingFillDownfromAction):
                command = f'fill_downfrom {r} {g} {b} {action.percentage} {action.wait}'
        elif isinstance(action, RgbRingSetEnabledAction):
            command = 'set_enabled ' + str(int(action.enabled))
        elif isinstance(action, RgbRingSetBrightnessAction):
            if 0 <= action.brightness <= 1:
                command = f'set_brightness {action.brightness}'
        elif isinstance(action, RgbRingBlankAction):
            command = 'blank'
        elif isinstance(action, RgbRingRainbowAction):
            command = f'rainbow {action.rounds} {action.wait}'

        if not command:
            return state

        return CompleteReducerResult(
            state=state,
            events=[RgbRingCommandEvent(command=command.split())],
        )

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import asyncio
import contextlib
import threading
from typing import TYPE_CHECKING, Callable, Coroutine, TypeVarTuple, Unpack

from redux.basic_types import FinishEvent
from typing_extensions import TypeVar

from ubo_app.constants import DEBUG_MODE

if TYPE_CHECKING:
    from asyncio import Future, Handle
    from asyncio.tasks import Task

    from redux.basic_types import TaskCreatorCallback


T = TypeVar('T', infer_variance=True)
Ts = TypeVarTuple('Ts')


class WorkerThread(threading.Thread):
    def __init__(self: WorkerThread) -> None:
        super().__init__()
        self.loop = asyncio.new_event_loop()
        if DEBUG_MODE:
            self.loop.set_debug(enabled=True)

    def run(self: WorkerThread) -> None:
        asyncio.set_event_loop(self.loop)

        from ubo_app.store import subscribe_event

        subscribe_event(FinishEvent, self.stop)
        self.loop.run_forever()

    def run_task(
        self: WorkerThread,
        task: Coroutine,
        callback: Callable[[Task], None] | None = None,
    ) -> Handle:
        def task_wrapper() -> None:
            result = self.loop.create_task(task)
            if callback:
                callback(result)

        return self.loop.call_soon_threadsafe(task_wrapper)

    def run_in_executor(
        self: WorkerThread,
        executor: object,
        task: Callable[[Unpack[Ts]], T],
        *args: Unpack[Ts],
    ) -> Future[T]:
        return self.loop.run_in_executor(executor, task, *args)

    async def shutdown(self: WorkerThread) -> None:
        while True:
            tasks = [
                t
                for t in asyncio.all_tasks(self.loop)
                if t is not asyncio.current_task(self.loop)
            ]
            if not tasks:
                break
            for task in tasks:
                with contextlib.suppress(asyncio.CancelledError, asyncio.TimeoutError):
                    await asyncio.wait_for(task, 0.1)
        self.loop.stop()

    def stop(self: WorkerThread) -> None:
        self.loop.call_soon_threadsafe(lambda: self.loop.create_task(self.shutdown()))


thread = WorkerThread()
thread.start()


def _create_task(
    task: Coroutine,
    callback: TaskCreatorCallback | None = None,
) -> Handle:
    return thread.run_task(task, callback)


def _run_in_executor(
    executer: object,
    task: Callable[[Unpack[Ts]], T],
    *args: Unpack[Ts],
) -> Future[T]:
    return thread.run_in_executor(executer, task, *args)


_ = _create_task, _run_in_executor



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Sequence

from kivy.clock import Clock
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.stencilview import StencilView
from kivy.uix.widget import Widget
from ubo_gui.app import UboApp

from ubo_app.store import autorun

if TYPE_CHECKING:
    from ubo_app.store.status_icons import IconState


class MenuAppFooter(UboApp):
    def _set_temperature_value(self: MenuAppFooter, value: float | None = None) -> None:
        if value is None:
            self.temperature.text = '-'
        else:
            self.temperature.text = f'{value:0.1f}°C'

    @cached_property
    def temperature_widget(self: MenuAppFooter) -> BoxLayout:
        layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
            size_hint=(None, 1),
        )

        self.temperature = Label(font_size=dp(14), size_hint=(None, 1), valign='middle')
        self.temperature.bind(
            texture_size=lambda temperature, texture_size: setattr(
                temperature,
                'width',
                texture_size[0],
            )
            or setattr(layout, 'width', temperature.width + dp(12)),
        )

        autorun(lambda state: state.sensors.temperature.value)(
            self._set_temperature_value,
        )

        icon = Label(
            text='device_thermostat',
            color='#ffffff',
            font_name='material_symbols',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(12),
        )
        layout.add_widget(icon)
        layout.add_widget(self.temperature)

        return layout

    def _set_light_value(self: MenuAppFooter, value: float | None = None) -> None:
        if value is None:
            self.light.color = (0.5, 0, 0, 1)
        else:
            v = min(value, 140) / 140
            self.light.color = (1, 1, 1, v)

    @cached_property
    def light_widget(self: MenuAppFooter) -> Label:
        self.light = Label(
            text='light_mode',
            color='#ffffff',
            font_name='material_symbols',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(16),
        )
        self.light.bind(
            texture_size=lambda light, texture_size: setattr(
                light,
                'width',
                texture_size[0],
            ),
        )

        autorun(lambda state: state.sensors.light.value)(self._set_light_value)

        return self.light

    @cached_property
    def clock_widget(self: MenuAppFooter) -> Label:
        clock = Label(font_size=dp(20), size_hint=(None, 1), valign='middle')
        clock.bind(
            texture_size=lambda clock, texture_size: setattr(
                clock,
                'width',
                texture_size[0],
            ),
        )
        local_timzone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

        def now() -> datetime.datetime:
            return datetime.datetime.now(local_timzone)

        def set_value(_: int | None = None) -> None:
            clock.text = now().strftime('%H:%M')

        def update(_: int | None = None) -> None:
            set_value()
            now_ = now()
            Clock.schedule_once(
                update,
                60 - now_.second - now_.microsecond / 1000000 + 0.05,
            )

        update()

        return clock

    def _render_icons(
        self: MenuAppFooter,
        selector_result: Sequence[IconState],
    ) -> None:
        icons = selector_result
        self.icons_layout.clear_widgets()
        for icon in list(reversed(icons))[:4]:
            label = Label(
                text=icon.symbol,
                color=icon.color,
                font_name='material_symbols',
                font_size=dp(20),
                font_features='fill=0',
                size_hint=(None, 1),
                width=dp(22),
            )
            self.icons_layout.add_widget(label)
        self.icons_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.icons_layout.bind(minimum_width=self.icons_layout.setter('width'))

    def _handle_depth_change(self: MenuAppFooter, _: Sequence[str]) -> None:
        is_fullscreen = False
        if not is_fullscreen:
            if self.normal_footer_layout in self.footer_layout.children:
                self.footer_layout.remove_widget(self.normal_footer_layout)
                self.footer_layout.add_widget(self.home_footer_layout)
        elif self.home_footer_layout in self.footer_layout.children:
            self.footer_layout.remove_widget(self.home_footer_layout)
            self.footer_layout.add_widget(self.normal_footer_layout)

    def set_icons_layout_x(self: MenuAppFooter, *_: list[Any]) -> None:
        self.icons_layout.x = (
            self.icons_widget.x + self.icons_widget.width - self.icons_layout.width
        )

    @cached_property
    def footer(self: MenuAppFooter) -> Widget | None:
        self.footer_layout = BoxLayout()

        self.normal_footer_layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
        )
        self.normal_footer_layout.add_widget(
            Label(
                text='reply',
                font_name='material_symbols',
                font_size=dp(20),
                font_features='fill=1',
                size_hint=(None, 1),
            ),
        )
        self.normal_footer_layout.add_widget(Widget(size_hint=(1, 1)))

        self.home_footer_layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
        )

        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.clock_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.temperature_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        self.home_footer_layout.add_widget(self.light_widget)
        self.home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))

        self.icons_widget = StencilView(size_hint=(1, 1))
        self.home_footer_layout.add_widget(self.icons_widget)

        self.icons_layout = BoxLayout(
            orientation='horizontal',
            spacing=dp(4),
            padding=0,
        )
        self.icons_widget.add_widget(self.icons_layout)

        self.icons_widget.bind(
            width=self.set_icons_layout_x,
            x=self.set_icons_layout_x,
            height=self.icons_layout.setter('height'),
            y=self.icons_layout.setter('y'),
        )
        self.icons_layout.bind(
            width=self.set_icons_layout_x,
            x=self.set_icons_layout_x,
        )

        autorun(lambda state: state.status_icons.icons)(self._render_icons)

        autorun(lambda state: state.main.path)(self._handle_depth_change)

        self.footer_layout.add_widget(self.home_footer_layout)

        return self.footer_layout



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import replace
from distutils.util import strtobool
from typing import cast

from constants import WIFI_STATE_ICON_ID, WIFI_STATE_ICON_PRIORITY, get_signal_icon
from redux import (
    BaseAction,
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.camera import (
    CameraBarcodeAction,
    CameraStopViewfinderAction,
)
from ubo_app.store.services.wifi import (
    GlobalWiFiState,
    WiFiAction,
    WiFiConnection,
    WiFiCreateEvent,
    WiFiEvent,
    WiFiState,
    WiFiType,
    WiFiUpdateAction,
    WiFiUpdateRequestAction,
    WiFiUpdateRequestEvent,
)
from ubo_app.store.status_icons import StatusIconsRegisterAction


def reducer(
    state: WiFiState | None,
    action: CameraBarcodeAction | WiFiAction,
) -> ReducerResult[WiFiState, BaseAction, WiFiEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return CompleteReducerResult(
                state=WiFiState(
                    connections=[],
                    state=GlobalWiFiState.UNKNOWN,
                    current_connection=None,
                ),
                actions=[WiFiUpdateRequestAction()],
            )
        raise InitializationActionError(action)

    if isinstance(action, CameraBarcodeAction):
        ssid = action.match.get('SSID')
        if ssid is None:
            return state

        return CompleteReducerResult(
            state=state,
            actions=[CameraStopViewfinderAction()],
            events=[
                WiFiCreateEvent(
                    connection=WiFiConnection(
                        ssid=ssid,
                        password=action.match.get('Password'),
                        type=cast(WiFiType, action.match.get('Type')),
                        hidden=strtobool(
                            action.match.get('Hidden') or 'false',
                        )
                        == 1,
                    ),
                ),
            ],
        )

    if isinstance(action, WiFiUpdateRequestAction):
        return CompleteReducerResult(
            state=replace(state, connections=None) if action.reset else state,
            events=[WiFiUpdateRequestEvent()],
        )

    if isinstance(action, WiFiUpdateAction):
        return CompleteReducerResult(
            state=replace(
                state,
                connections=action.connections,
                state=action.state,
                current_connection=action.current_connection,
            ),
            actions=[
                StatusIconsRegisterAction(
                    icon={
                        GlobalWiFiState.CONNECTED: get_signal_icon(
                            action.current_connection.signal_strength
                            if action.current_connection
                            else 0,
                        ),
                        GlobalWiFiState.DISCONNECTED: 'signal_wifi_off',
                        GlobalWiFiState.PENDING: 'wifi_find',
                        GlobalWiFiState.NEEDS_ATTENTION: (
                            'signal_wifi_statusbar_not_connected'
                        ),
                        GlobalWiFiState.UNKNOWN: 'perm_scan_wifi',
                    }[action.state],
                    priority=WIFI_STATE_ICON_PRIORITY,
                    id=WIFI_STATE_ICON_ID,
                ),
            ],
        )

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from enum import StrEnum

from redux import BaseAction, BaseEvent


class Key(StrEnum):
    BACK = 'Back'
    HOME = 'Home'
    UP = 'Up'
    DOWN = 'Down'
    L1 = 'L1'
    L2 = 'L2'
    L3 = 'L3'


class KeypadAction(BaseAction):
    key: Key


class KeypadKeyUpAction(KeypadAction):
    ...


class KeypadKeyDownAction(KeypadAction):
    ...


class KeypadKeyPressAction(KeypadAction):
    ...


class KeypadEvent(BaseEvent):
    key: Key


class KeypadKeyPressEvent(KeypadEvent):
    ...



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import pathlib
import weakref
from functools import cached_property
from typing import TYPE_CHECKING, Sequence

from debouncer import DebounceOptions, debounce
from kivy.clock import Clock, mainthread
from kivy.lang.builder import Builder
from redux import EventSubscriptionOptions
from ubo_gui.app import UboApp
from ubo_gui.gauge import GaugeWidget
from ubo_gui.menu import MenuWidget
from ubo_gui.notification import NotificationWidget
from ubo_gui.page import PageWidget
from ubo_gui.volume import VolumeWidget

from ubo_app.store.main import SetMenuPathAction
from ubo_app.store.services.keypad import Key, KeypadKeyPressEvent
from ubo_app.store.services.notifications import (
    NotificationDisplayType,
    NotificationsClearAction,
    NotificationsDisplayEvent,
)

from .store import autorun, dispatch, subscribe_event

if TYPE_CHECKING:
    from kivy.uix.widget import Widget
    from ubo_gui.menu.types import Item, Menu


class HomePage(PageWidget):
    def __init__(
        self: HomePage,
        items: Sequence[Item] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        super().__init__(items, *args, **kwargs)

        self.ids.central_column.add_widget(self.cpu_gauge)
        self.ids.central_column.add_widget(self.ram_gauge)

        self.volume_widget = VolumeWidget()
        self.ids.right_column.add_widget(self.volume_widget)

        autorun(lambda state: state.sound.playback_volume)(self._sync_output_volume)

    def _sync_output_volume(self: HomePage, selector_result: float) -> None:
        self.volume_widget.value = selector_result * 100

    @cached_property
    def cpu_gauge(self: HomePage) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(
            value=psutil.cpu_percent(percpu=False),
            fill_color='#24D636',
            label='CPU',
        )

        def set_value(_: int) -> None:
            gauge.value = psutil.cpu_percent(percpu=False)

        Clock.schedule_interval(set_value, 1)

        return gauge

    @cached_property
    def ram_gauge(self: HomePage) -> GaugeWidget:
        import psutil

        gauge = GaugeWidget(
            value=psutil.virtual_memory().percent,
            fill_color='#D68F24',
            label='RAM',
        )

        def set_value(_: int) -> None:
            gauge.value = psutil.virtual_memory().percent

        Clock.schedule_interval(set_value, 1)

        return gauge


class MenuWidgetWithHomePage(MenuWidget):
    def render_items(self: MenuWidgetWithHomePage, *_: object) -> None:
        if self.depth == 1:
            self.current_screen = HomePage(self.current_menu_items, name='Page 1 0')
        else:
            super().render_items()


def set_path(menu_widget: MenuWidget, _: list[tuple[Menu, int] | PageWidget]) -> None:
    dispatch(
        SetMenuPathAction(
            path=[i.title for i in menu_widget.stack],
        ),
    )


class MenuAppCentral(UboApp):
    def __init__(self: MenuAppCentral, **kwargs: object) -> None:
        super().__init__(**kwargs)
        self.menu_widget = MenuWidgetWithHomePage()

        _self = weakref.ref(self)

        @autorun(lambda state: state.main.menu)
        @debounce(0.1, DebounceOptions(leading=True, trailing=True, time_window=0.1))
        async def _(menu: Menu | None) -> None:
            self = _self()
            if not self or not menu:
                return
            mainthread(self.menu_widget.set_root_menu)(menu)

    def handle_title_change(self: MenuAppCentral, _: MenuWidget, title: str) -> None:
        self.root.title = title

    @cached_property
    def central(self: MenuAppCentral) -> Widget | None:
        """Build the main menu and initiate it."""
        self.root.title = self.menu_widget.title
        self.menu_widget.bind(title=self.handle_title_change)
        self.menu_widget.bind(current_screen=set_path)

        subscribe_event(
            KeypadKeyPressEvent,
            self.handle_key_press_event,
            options=EventSubscriptionOptions(immediate_run=True),
        )

        subscribe_event(
            NotificationsDisplayEvent,
            self.display_notification,
            options=EventSubscriptionOptions(immediate_run=True),
        )

        return self.menu_widget

    def handle_key_press_event(
        self: MenuAppCentral,
        key_press_event: KeypadKeyPressEvent,
    ) -> None:
        if key_press_event.key == Key.L1:
            self.menu_widget.select(0)
        if key_press_event.key == Key.L2:
            self.menu_widget.select(1)
        if key_press_event.key == Key.L3:
            self.menu_widget.select(2)
        if key_press_event.key == Key.BACK:
            self.menu_widget.go_back()
        if key_press_event.key == Key.UP:
            self.menu_widget.go_up()
        if key_press_event.key == Key.DOWN:
            self.menu_widget.go_down()

    @mainthread
    def display_notification(
        self: MenuAppCentral,
        event: NotificationsDisplayEvent,
    ) -> None:
        notification = event.notification
        application = NotificationWidget(
            title='Notification',
            notification_title=notification.title,
            content=notification.content,
            icon=notification.icon,
            color=notification.color,
        )

        application.bind(
            on_dismiss=lambda _: (
                application.dispatch('on_close'),
                dispatch(
                    NotificationsClearAction(notification=notification),
                ),
            ),
        )

        self.menu_widget.open_application(application)

        if notification.display_type is NotificationDisplayType.FLASH:
            Clock.schedule_once(lambda _: application.dispatch('on_close'), 4)


Builder.load_file(
    pathlib.Path(__file__).parent.joinpath('home_page.kv').resolve().as_posix(),
)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from setup import init_service

from ubo_app.load_services import register_service
from ubo_app.utils import IS_RPI

register_service(
    service_id='keypad',
    label='Keypad',
)

if IS_RPI:
    init_service()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os
from distutils.util import strtobool
from pathlib import Path

import dotenv

dotenv.load_dotenv(Path(__file__).parent / '.env')

USERNAME = os.environ.get('UBO_USERNAME', 'ubo')
INSTALLATION_PATH = os.environ.get('UBO_INSTALLATION_PATH', '/opt/ubo')
DEBUG_MODE = strtobool(os.environ.get('UBO_DEBUG', 'False')) == 1
LOG_LEVEL = os.environ.get('UBO_LOG_LEVEL', 'DEBUG' if DEBUG_MODE else None)
GUI_LOG_LEVEL = os.environ.get('UBO_GUI_LOG_LEVEL', 'DEBUG' if DEBUG_MODE else None)
SERVICES_PATH = (
    os.environ.get('UBO_SERVICES_PATH', '').split(':')
    if os.environ.get('UBO_SERVICES_PATH')
    else []
)
SOCKET_PATH = Path('/run/ubo').joinpath('system_manager.sock').as_posix()

DEBUG_MODE_DOCKER = strtobool(os.environ.get('UBO_DEBUG_DOCKER', 'False')) == 1
DOCKER_PREFIX = os.environ.get('UBO_DOCKER_PREFIX', '')
DOCKER_INSTALLATION_LOCK_FILE = Path('/var/run/ubo/docker_installation.lock')



# ruff: noqa: D100, D101, D102, D103, D104, D107
import os
import sys
import threading
import traceback
from pathlib import Path
from types import TracebackType

import dotenv
from redux import FinishAction

dotenv.load_dotenv(Path(__file__).parent / '.dev.env')


def setup_logging() -> None:
    from ubo_app.constants import GUI_LOG_LEVEL, LOG_LEVEL

    if LOG_LEVEL:
        import logging

        import ubo_app.logging

        level = getattr(
            ubo_app.logging,
            LOG_LEVEL,
            getattr(logging, LOG_LEVEL, logging.INFO),
        )

        ubo_app.logging.logger.setLevel(level)
        ubo_app.logging.add_file_handler(ubo_app.logging.logger, level)
        ubo_app.logging.add_stdout_handler(ubo_app.logging.logger, level)
    if GUI_LOG_LEVEL:
        import logging

        import ubo_gui.logger

        level = getattr(
            ubo_gui.logger,
            GUI_LOG_LEVEL,
            getattr(logging, GUI_LOG_LEVEL, logging.INFO),
        )

        ubo_gui.logger.logger.setLevel(level)
        ubo_gui.logger.add_file_handler(level)
        ubo_gui.logger.add_stdout_handler(level)


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    setup_logging()

    if len(sys.argv) > 1 and sys.argv[1] == 'bootstrap':
        from ubo_app.system.bootstrap import bootstrap

        bootstrap(
            with_docker='--with-docker' in sys.argv,
            for_packer='--for-packer' in sys.argv,
        )
        sys.exit(0)

    def global_exception_handler(
        exception_type: type,
        value: int,
        tb: TracebackType,
    ) -> None:
        from ubo_app.logging import logger

        logger.error(f'Uncaught exception: {exception_type}: {value}')
        logger.error(''.join(traceback.format_tb(tb)))

    # Set the global exception handler
    sys.excepthook = global_exception_handler

    def thread_exception_handler(args: threading.ExceptHookArgs) -> None:
        import traceback

        from ubo_app.logging import logger

        logger.error(
            f"""Exception in thread {args.thread.name if args.thread else "-"}: {
            args.exc_type} {args.exc_value}""",
        )
        logger.error(''.join(traceback.format_tb(args.exc_traceback)))

    threading.excepthook = thread_exception_handler

    os.environ['KIVY_NO_CONFIG'] = '1'
    os.environ['KIVY_NO_FILELOG'] = '1'

    import headless_kivy_pi.config

    headless_kivy_pi.config.setup_headless_kivy({'automatic_fps': True})

    from kivy.clock import Clock

    from ubo_app.load_services import load_services
    from ubo_app.menu import MenuApp

    load_services()
    app = MenuApp()

    try:
        app.run()
    finally:
        from ubo_app.store import dispatch

        dispatch(FinishAction())

        # Needed since redux is scheduled using Clock scheduler and Clock doesn't run
        # after app is stopped.
        Clock.tick()


if __name__ == '__main__':
    main()



# ruff: noqa: D100, D101, D102, D103, D104, D107, PLR2004
from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING, Sequence, cast

import board
import neopixel
from adafruit_blinka.microcontroller.generic_micropython import Pin
from ubo_gui.menu import warnings

from ubo_app.logging import add_file_handler, add_stdout_handler, get_logger

if TYPE_CHECKING:
    from ubo_app.store.services.rgb_ring import Color

BRIGHTNESS = 1.0
NUM_LEDS = 27
ORDER = neopixel.GRB


class LEDManager:
    def __init__(self: LEDManager) -> None:
        self.logger = get_logger('led-manager')
        add_file_handler(self.logger, logging.DEBUG)
        add_stdout_handler(self.logger, logging.DEBUG)
        self.logger.setLevel(logging.DEBUG)
        self.logger.debug('Initialising LEDManager...')

        self.led_ring_present = True
        self.current_color = None
        self._stop: bool = False
        self.current_bright_one = 1
        self.brightness = BRIGHTNESS
        if self.brightness < 0 or self.brightness > 1:
            warnings.warn(
                'Invalid brightness value in config file',
                Warning,
                stacklevel=2,
            )
            self.brightness = 1.0

        self.num_leds = NUM_LEDS
        self.pixels = neopixel.NeoPixel(
            pin=cast(Pin, board.D12),
            n=self.num_leds,
            brightness=1,
            bpp=3,
            auto_write=False,
            pixel_order=ORDER,
        )

    def stop(self: LEDManager) -> None:
        self._stop = True

    def adjust_brightness(
        self: LEDManager,
        color: Color,
    ) -> Color:
        b = self.brightness
        return (color[0] * b, color[1] * b, color[2] * b)

    def set_brightness(self: LEDManager, b: float) -> None:
        self.brightness = b

    def set_enabled(self: LEDManager, *, enabled: bool = True) -> None:
        if enabled is False:
            self.blank()
        self.led_ring_present = enabled

    def set_all(self: LEDManager, color: Color) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        self.pixels.fill(color)
        self.pixels.show()

    def blank(self: LEDManager) -> None:
        if not self.led_ring_present:
            return
        self.set_all((0, 0, 0))

    def fill_upto(
        self: LEDManager,
        color: Color,
        percentage: float,
        wait: float,
    ) -> None:
        if not self.led_ring_present:
            return
        for i in range(int(self.num_leds * percentage)):
            if self._stop is True:
                self.blank()
                return
            self.pixels[i] = self.adjust_brightness(color)
            time.sleep(wait / 1000)
            self.pixels.show()
        time.sleep(5 * wait / 1000)

    def fill_downfrom(
        self: LEDManager,
        color: Color,
        percentage: float,
        wait: float,
    ) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        self.pixels[: int(self.num_leds * percentage)] = [color] * int(
            self.num_leds * percentage,
        )
        self.pixels.show()
        time.sleep(5 * wait / 1000)
        for i in range(int(self.num_leds * percentage) - 1, -1, -1):
            if self._stop is True:
                self.blank()
                return
            self.pixels[i] = (0, 0, 0)
            time.sleep(wait / 1000)
            self.pixels.show()

    def progress_wheel_step(self: LEDManager, color: Color) -> None:
        if not self.led_ring_present:
            return
        dim_factor = 20
        color = self.adjust_brightness(color)
        self.set_all(
            (color[0] / dim_factor, color[1] / dim_factor, color[2] / dim_factor),
        )
        self.current_bright_one = (self.current_bright_one + 1) % self.num_leds
        before = (self.current_bright_one - 1) % self.num_leds
        if before < 0:
            before = self.num_leds + before
        after = (self.current_bright_one + 1) % self.num_leds

        self.pixels[before] = color
        self.pixels[after] = color
        self.pixels[self.current_bright_one] = color
        self.pixels.show()

    def wheel(self: LEDManager, pos: int) -> Color:
        # Input a value 0 to 255 to get a color value.
        # The colours are a transition r - g - b - back to r.
        if pos < 0 or pos > 0b11111111:
            r = g = b = 0
        elif pos < 0b01010101:
            r = int(pos * 3)
            g = int(0b11111111 - pos * 3)
            b = 0
        elif pos < 0b10101010:
            pos -= 85
            r = int(0b11111111 - pos * 3)
            g = 0
            b = int(pos * 3)
        else:
            pos -= 0b10101010
            r = 0
            g = int(pos * 3)
            b = int(0b11111111 - pos * 3)
        return (r, g, b) if ORDER in (neopixel.RGB, neopixel.GRB) else (r, g, b, 0)

    # wait is in milliseconds
    def rainbow(
        self: LEDManager,
        rounds: int = 50,
        wait: float = 1,
    ) -> None:
        if not self.led_ring_present:
            return
        for _ in range(rounds):
            for j in range(255):
                if self._stop is True:
                    self.blank()
                    return
                for i in range(self.num_leds):
                    pixel_index = (i * 256 // self.num_leds) + j
                    (r, g, b, *_) = self.wheel(pixel_index & 255)
                    self.pixels[i] = (
                        r * self.brightness,
                        g * self.brightness,
                        b * self.brightness,
                    )
                self.pixels.show()
                time.sleep(wait / 1000)
        self.blank()

    def pulse(self: LEDManager, color: Color, wait: float, repetitions: int) -> None:
        # wait is in milliseconds
        # repetitions is the number of retepting pluses
        if not self.led_ring_present:
            return
        dim_steps = 10
        color = self.adjust_brightness(color)
        for _ in range(repetitions):
            for i in range(dim_steps):
                if self._stop is True:
                    self.blank()
                    return
                m = i / dim_steps
                self.pixels.fill((color[0] * m, color[1] * m, color[2] * m))
                self.pixels.show()
                time.sleep(wait / 10000)
            for i in range(1, dim_steps):
                if self._stop is True:
                    self.blank()
                    return
                j = (dim_steps - i) / dim_steps
                self.pixels.fill((color[0] * j, color[1] * j, color[2] * j))
                self.pixels.show()
                time.sleep(wait / 10000)
        self.blank()

    def blink(self: LEDManager, color: Color, wait: float, repetitions: int) -> None:
        # wait is in milliseconds
        # repetitions is the number of blinks
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        for _ in range(repetitions):
            if self._stop is True:
                self.blank()
                return
            self.pixels.fill((color[0], color[1], color[2]))
            self.pixels.show()
            time.sleep(wait / 1000)
            self.blank()
            time.sleep(1.5 * wait / 1000)

    def spinning_wheel(
        self: LEDManager,
        color: Color,
        wait: float = 1,
        length: int = 5,
        repetitions: int = 5,
    ) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        ring: list[Color] = [(0, 0, 0)] * self.num_leds
        ring[0:length] = [color] * (length)
        if length > self.num_leds:
            warnings.warn(
                f'invalid light strip length! must be under {self.num_leds}',
                Warning,
                stacklevel=2,
            )
            return
        for _ in range(repetitions):
            for i in range(self.num_leds):
                if self._stop is True:
                    self.blank()
                    return
                shifted = ring[i:] + ring[:i]
                # for j in self.num_leds
                self.pixels[:] = shifted
                self.pixels.show()
                time.sleep(wait / 1000)
        self.blank()

    def progress_wheel(self: LEDManager, color: Color, percentage: float) -> None:
        # percentage is a float value between 0 and 1
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        ring: list[Color] = [(0, 0, 0)] * self.num_leds
        ring[0 : int(self.num_leds * percentage)] = [color] * (
            int(self.num_leds * percentage)
        )
        self.pixels[:] = ring
        self.pixels.show()

    def run_command(self: LEDManager, incoming: Sequence[str]) -> None:  # noqa: C901, PLR0912
        self.logger.info('Executing LED command', extra={'incoming': incoming})
        self.incoming = incoming
        self._stop = False
        if incoming[0] == 'set_enabled' and len(incoming) == 2:
            self.set_enabled(enabled=incoming[1] == '1')
        # set brightness of LEDs
        if incoming[0] == 'set_brightness' and len(incoming) == 2:
            brightness_value = float(incoming[1])
            if 0 < brightness_value <= 1:
                self.set_brightness(brightness_value)
        if incoming[0] == 'set_all' and len(incoming) == 4:
            self.set_all((int(incoming[1]), int(incoming[2]), int(incoming[3])))
        if incoming[0] == 'blank':
            self.blank()
        if incoming[0] == 'rainbow' and len(incoming) == 3:
            self.rainbow(int(incoming[1]), float(incoming[2]))
        if incoming[0] == 'pulse' and len(incoming) == 6:
            # pulse(self, color, wait, repetitions):
            self.pulse(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                wait=int(incoming[4]),
                repetitions=int(incoming[5]),
            )
        if incoming[0] == 'blink' and len(incoming) == 6:
            # blink(self, color, wait, repetitions):
            self.blink(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                wait=int(incoming[4]),
                repetitions=int(incoming[5]),
            )
        if incoming[0] == 'progress_wheel_step' and len(incoming) == 4:
            self.progress_wheel_step(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
            )
        if incoming[0] == 'spinning_wheel':  # noqa: SIM102
            # spinning_wheel(self, color, wait, length, repetitions):
            if len(incoming) == 7:
                self.spinning_wheel(
                    (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                    wait=int(incoming[4]),
                    length=int(incoming[5]),
                    repetitions=int(incoming[6]),
                )
        if incoming[0] == 'progress_wheel' and len(incoming) == 5:
            # progress_wheel(self, color, percentage):
            self.progress_wheel(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
            )
        if incoming[0] == 'fill_upto' and len(incoming) == 6:
            # fill_upto(self, color, percentage, wait):
            self.fill_upto(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
                wait=int(incoming[5]),
            )
        if incoming[0] == 'fill_downfrom' and len(incoming) == 6:
            # fill_upto(self, color, percentage, wait):
            self.fill_downfrom(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
                wait=int(incoming[5]),
            )
        self._stop = False



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

from types import ModuleType
from typing import Any, Generator, Iterator, Self, cast

from ubo_app.logging import logger


class Fake(ModuleType):
    def __init__(self: Fake, *args: object, **kwargs: object) -> None:
        logger.verbose('Initializing `Fake`', extra={'args_': args, 'kwargs': kwargs})
        self.iterated = False
        super().__init__('')

    def __init_subclass__(cls: type[Fake], **kwargs: dict[str, Any]) -> None:
        logger.verbose('Subclassing `Fake`', extra={'cls': cls, 'kwargs': kwargs})

    def __getattr__(self: Fake, attr: str) -> Fake:
        logger.verbose(
            'Accessing fake attribute of a `Fake` insta',
            extra={'attr': attr},
        )
        if attr == '__file__':
            return cast(Fake, 'fake')
        return self

    def __getitem__(self: Fake, key: object) -> Fake:
        logger.verbose(
            'Accessing fake item of a `Fake` instance',
            extra={'key': key},
        )
        return self

    def __call__(self: Fake, *args: object, **kwargs: dict[str, Any]) -> Fake:
        logger.verbose(
            'Calling a `Fake` instance',
            extra={'args_': args, 'kwargs': kwargs},
        )
        return self

    def __await__(self: Fake) -> Generator[Fake | None, Any, Any]:
        yield
        return Fake()

    def __next__(self: Fake) -> Fake:
        if self.iterated:
            raise StopIteration
        self.iterated = True
        return self

    def __anext__(self: Fake) -> Fake:
        if self.iterated:
            raise StopAsyncIteration
        self.iterated = True
        return self

    def __iter__(self: Fake) -> Iterator[Fake]:
        return Fake()

    def __aiter__(self: Fake) -> Iterator[Fake]:
        return Fake()

    def __enter__(self: Fake) -> Fake:  # noqa: PYI034
        return self

    def __exit__(self: Fake, *_: object) -> None:
        pass

    async def __aenter__(self: Self) -> Self:
        return self

    async def __aexit__(self: Fake, *_: object) -> None:
        pass

    def __mro_entries__(self: Fake, bases: tuple[type[Fake]]) -> tuple[type[Fake]]:
        logger.verbose(
            'Getting MRO entries of a `Fake` instance',
            extra={'bases': bases},
        )
        return (cast(type, self),)

    def __len__(self: Fake) -> int:
        return 1

    def __index__(self: Fake) -> int:
        return 1

    def __contains__(self: Fake, _: object) -> bool:
        return True

    def __eq__(self: Fake, _: object) -> bool:
        return True

    def __ne__(self: Fake, _: object) -> bool:
        return False

    def __str__(self: Fake) -> str:
        return 'Fake'

    def __repr__(self: Fake) -> str:
        return 'Fake'



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from enum import StrEnum

from immutable import Immutable
from redux import BaseAction, BaseEvent


class SoundDevice(StrEnum):
    INPUT = 'Input'
    OUTPUT = 'Output'


class SoundAction(BaseAction):
    ...


class SoundSetVolumeAction(SoundAction):
    volume: float
    device: SoundDevice


class SoundChangeVolumeAction(SoundAction):
    amount: float
    device: SoundDevice


class SoundSetMuteStatusAction(SoundAction):
    mute: bool
    device: SoundDevice


class SoundToggleMuteStatusAction(SoundAction):
    device: SoundDevice


class SoundPlayChimeAction(SoundAction):
    name: str


class SoundEvent(BaseEvent):
    ...


class SoundPlayChimeEvent(SoundEvent):
    name: str


class SoundState(Immutable):
    playback_volume: float
    is_playback_mute: bool
    capture_volume: float
    is_capture_mute: bool



"""Docker reducer."""
from __future__ import annotations

from dataclasses import field, replace

from immutable import Immutable
from redux import (
    BaseEvent,
    CombineReducerAction,
    CombineReducerInitAction,
    InitAction,
    InitializationActionError,
    ReducerResult,
    combine_reducers,
)

from ubo_app.constants import DEBUG_MODE_DOCKER, DOCKER_PREFIX
from ubo_app.store.services.docker import (
    DockerAction,
    DockerImageAction,
    DockerImageEvent,
    DockerImageSetDockerIdAction,
    DockerImageSetStatusAction,
    DockerServiceState,
    DockerSetStatusAction,
    DockerState,
    ImageState,
)
from ubo_app.store.services.ip import IpUpdateAction

Action = InitAction | DockerAction


def service_reducer(
    state: DockerServiceState | None,
    action: Action,
) -> ReducerResult[DockerServiceState, Action, BaseEvent]:
    """Docker reducer."""
    if state is None:
        if isinstance(action, InitAction):
            return DockerServiceState()
        raise InitializationActionError(action)

    if isinstance(action, DockerSetStatusAction):
        return replace(state, status=action.status)

    return state


class ImageEntry(Immutable):
    """An image to be used in a Docker container."""

    id: str
    label: str
    icon: str
    path: str
    ports: dict[str, str] = field(default_factory=dict)
    hosts: dict[str, str] = field(default_factory=dict)
    volumes: list[str] | None = None


IMAGES = {
    image.id: image
    for image in [
        ImageEntry(
            id='home_assistant',
            label='Home Assistant',
            icon='home',
            path=DOCKER_PREFIX + 'homeassistant/home-assistant:stable',
            ports={'8123/tcp': '8123'},
        ),
        ImageEntry(
            id='home_bridge',
            label='Home Bridge',
            icon='home_work',
            path=DOCKER_PREFIX + 'homebridge/homebridge:latest',
        ),
        ImageEntry(
            id='portainer',
            label='Portainer',
            icon='settings_applications',
            path=DOCKER_PREFIX + 'portainer/portainer-ce:latest',
            volumes=['/var/run/docker.sock:/var/run/docker.sock'],
        ),
        ImageEntry(
            id='pi_hole',
            label='Pi-hole',
            icon='dns',
            path=DOCKER_PREFIX + 'pihole/pihole:latest',
        ),
        ImageEntry(
            id='ollama',
            label='Ollama',
            icon='smart_toy',
            path=DOCKER_PREFIX + 'ollama/ollama:latest',
            ports={'11434/tcp': '11434'},
        ),
        ImageEntry(
            id='open_webui',
            label='Open WebUI',
            icon='code',
            path=DOCKER_PREFIX + 'ghcr.io/open-webui/open-webui:main',
            ports={'8080/tcp': '8080'},
            hosts={'host.docker.internal': 'ollama'},
        ),
        *(
            [
                ImageEntry(
                    id='alpine',
                    label='Alpine',
                    icon='code',
                    path=DOCKER_PREFIX + 'alpine:latest',
                ),
            ]
            if DEBUG_MODE_DOCKER
            else []
        ),
    ]
}
IMAGE_IDS = list(IMAGES.keys())


def image_reducer(
    state: ImageState | None,
    action: DockerImageAction | CombineReducerAction | IpUpdateAction,
) -> ImageState:
    """Image reducer."""
    if state is None:
        if isinstance(action, CombineReducerInitAction):
            image = IMAGES[action.key]
            return ImageState(
                id=image.id,
                label=image.label,
                icon=image.icon,
                path=image.path,
            )
        raise InitializationActionError(action)

    if isinstance(action, IpUpdateAction):
        return replace(
            state,
            ip_addresses=[
                ip for interface in action.interfaces for ip in interface.ip_addresses
            ],
        )

    if not isinstance(action, DockerImageAction) or action.image != state.id:
        return state

    if isinstance(action, DockerImageSetStatusAction):
        return replace(
            state,
            status=action.status,
            ports=action.ports if action.ports else state.ports,
            container_ip=action.ip,
        )

    if isinstance(action, DockerImageSetDockerIdAction):
        return replace(state, docker_id=action.docker_id)

    return state


reducer, reducer_id = combine_reducers(
    state_type=DockerState,
    action_type=DockerImageAction,
    event_type=DockerImageEvent,
    service=service_reducer,
    **{image: image_reducer for image in IMAGE_IDS},
)
